import { useCallback } from "react";
import {
  BodyParams,
  Combination,
  ErrorCorrector,
  FetchConfig,
  Fetcher,
  HttpMethod,
  RequestFactory,
  RequestIntercepter,
  ResponseTransformer,
  SearchParams,
} from "./types";
import { useCreateBody } from "./useCreateBody";
import { useCreateHeaders } from "./useCreateHeaders";
import { useCreateRequest } from "./useCreateRequest";
import { useCreateUrl } from "./useCreateUrl";
import { useCreateErrorCorrector } from "./useCreateErrorCorrector";
import { useCreateFetchConfig } from "./useCreateFetchConfig";
import { useCreateFetcher } from "./useCreateFetcher";
import { useCreateRequestIntercepter } from "./useCreateRequestIntercepter";
import { useCreateResponseTransformer } from "./useCreateResponseTransformer";

export interface FetchOptionBase {
  body?: BodyParams;
  bodyCombination?: Combination;
  config?: FetchConfig;
  configCombination?: Combination;
  correct?: ErrorCorrector;
  correctCombination?: Combination;
  fetcher?: Fetcher;
  hasFiles?: boolean;
  headers?: HeadersInit;
  headersCombination?: Combination;
  intercept?: RequestIntercepter;
  interceptCombination?: Combination;
  searchParams?: SearchParams;
  searchParamsCombination?: Combination;
  transform?: ResponseTransformer;
}

export interface FetchVariables extends FetchOptionBase {
  url: string;
  method?: HttpMethod;
  signal?: AbortSignal;
}

export interface UseFetchOptions extends FetchOptionBase {
  baseUrl?: string;
  createRequest?: RequestFactory;
}

export type FetchFunction = <TData = any>(options: FetchVariables) => Promise<TData>;

export function useFetch(
  options: UseFetchOptions = {}
): FetchFunction {
  const createFetchConfig = useCreateFetchConfig(options);
  const createUrl = useCreateUrl(options);
  const createHeaders = useCreateHeaders(options);
  const createBody = useCreateBody(options);
  const createRequest = useCreateRequest(options);
  const createFetcher = useCreateFetcher(options);
  const createIntercepter = useCreateRequestIntercepter(options);
  const createTransformer = useCreateResponseTransformer(options);
  const createCorrector = useCreateErrorCorrector(options);

  return useCallback(
    async <TData>(options: FetchVariables): Promise<TData> => {
      let uri: URL;
      let body: BodyInit | null;
      let headers: HeadersInit | undefined;
      let config: FetchConfig;
      let intercept: RequestIntercepter;
      let transform: ResponseTransformer;
      let correct: ErrorCorrector;
      let request: Request;
      let response: Response | undefined;
      let fetch: Fetcher;

      return Promise.resolve()
        .then(() => uri = createUrl(options))
        .then(() => body = createBody(options))
        .then(() => headers = createHeaders(options))
        .then(() => config = createFetchConfig(options))
        .then(() => intercept = createIntercepter(options))
        .then(() => transform = createTransformer(options))
        .then(() => correct = createCorrector(options))
        .then(() => fetch = createFetcher(options))
        .then(() => createRequest(uri, config, headers, body, options.signal))
        .then(async req => await intercept(request = req))
        .then(async req => await fetch(request = req))
        .then(async res => await transform(request, response = res))
        .catch(async err => await correct(err, request, response));
    },
    [
      createFetchConfig,
      createUrl,
      createHeaders,
      createBody,
      createRequest,
      createFetcher,
      createIntercepter,
      createTransformer,
      createCorrector,
    ],
  );
}
