import isPlainObject from 'lodash/isPlainObject';
import isString from 'lodash/isString';
import isMap from 'lodash/isMap';
import isSet from 'lodash/isSet';
import isArray from 'lodash/isArray';
import isDate from 'lodash/isDate';
import isBoolean from 'lodash/isBoolean';
import isNaN from 'lodash/isNaN';
import isNumber from 'lodash/isNumber';
import { FormatOptions } from "./FetchContext";
import { BodyParams, Combination, SearchParams } from "./types";
import { FetchError } from './FetchError';

export const resolveFormatOptions = (
  options: FormatOptions | undefined,
): Required<FormatOptions> => ({
  bool: options?.bool ?? String,
  null: options?.null ?? '',
  nan: options?.nan  ?? 'NaN',
  date: options?.date  ?? (d => d.toISOString()),
  skipNull: options?.skipNull ?? true,
  skipEmptyString: options?.skipEmptyString ?? true,
  skipNaN: options?.skipNaN ??  true,
  arrayFormat: options?.arrayFormat ?? 'default',
  objectFormat: options?.objectFormat ?? 'bracket',
});

export function searchParamsIntoUrl(
  url: URL,
  params: SearchParams,
  options: FormatOptions & { combination: Combination },
): void {
  if (options.combination === "overrides") {
    url.search = ""; // clean
  }
  if (params === null || params === "") {
    return;
  } else if (typeof params === 'string') {
    params = new URLSearchParams();
  } else if (params instanceof URLSearchParams) {
    // nothing
  } else if (Array.isArray(params)) {
    for (const [key, value] of params) {
      url.searchParams.set(key, String(value));
    }
    return;
  }
  merge(
    url.searchParams as Merger,
    params,
    options,
  );
}

const isObject = <T = any>(v: any): v is Record<string, T> => {
  return isPlainObject(v);
};

export function mergeBodys(
  bodys: BodyParams[],
  hasFiles: boolean | undefined,
  options: FormatOptions & { combination: Combination },
): BodyParams | null {
  const size = bodys.length;
  if (size === 0) {
    return null;
  }
  if (size === 1 || options.combination === 'overrides') {
    const body = bodys[size - 1];
    if (isBodyInit(body) || body === null) {
      return body;
    }
    if (hasFiles) {
      const data = new FormData();
      merge(data, body, options);
      return data;
    }
    return body;
  }
  if (hasFiles || bodys.some(b => b instanceof FormData)) {
    const data = new FormData();
    bodys.forEach(body => merge(data, body, options));
    return data;
  }
  return bodys.reduce((acc, body) => {
    if (!body) {
      return acc;
    } else if (
      !acc ||
      body instanceof ReadableStream ||
      body instanceof Blob ||
      body instanceof ArrayBuffer ||
      ArrayBuffer.isView(body) ||
      body instanceof FormData
    ) {
      return body;
    } else if (acc instanceof URLSearchParams) {
      if (!(body instanceof URLSearchParams)) {
        throw new Error(
          `Should not merge a ${displayName(body)} into a URLSearchParams.`,
        );
      }
      body.forEach((_, name) => {
        body.getAll(name).forEach((value, index) => {
          if (index === 0) {
            acc.set(name, value); // replace
          } else {
            acc.append(name, value); // append
          }
        })
      })
    } else if (isObject(acc)) {
      if (body instanceof URLSearchParams) {
        body.forEach((_, name) => {
          const values = body.getAll(name);
          const value = values.length > 1 ? values : values[0];
          // todo expand the name to path
          // @ts-ignore
          acc[name] = value;
        })
      } else if (isObject(body)) {
        // merge first level only
        return {
          ...acc,
          ...body,
        }
      } else {
        throw new Error(
          `Should not merge a ${displayName(body)} into a plaint object.`,
        );
      }
    } else {
      const accName = displayName(acc);
      const bodyName = displayName(body);
      if (accName === bodyName) {
        throw new FetchError(
          `Should not merge a ${bodyName} into another ${acc}`,
        );
      }
      throw new FetchError(
        `Should not merge a ${bodyName} into a ${acc}`,
      );
    }
    return acc;
  });
}

export function buildBody(
  body: BodyParams,
  hasFiles: boolean,
  options: FormatOptions,
): BodyInit | null {
  if (isBodyInit(body) || body == null) {
    return body;
  }
  if (hasFiles) {
    const data = new FormData();
    merge(data, body, options);
    return data;
  }
  return JSON.stringify(body);
}

const displayName = (a: any) => Object.prototype.toString.call(a);

export const isBodyInit = (body: BodyParams): body is BodyInit => {
  return body != null && (
    body instanceof ReadableStream
    || body instanceof Blob
    || body instanceof ArrayBuffer
    || ArrayBuffer.isView(body)
    || body instanceof FormData
    || body instanceof URLSearchParams
    || typeof body === 'string'
  );
}

interface Merger {
  append: (path: string, value: Blob | string) => void;
  set: (path: string, value: Blob | string) => void;
}

const arrayKey = (
  strategy: FormatOptions['arrayFormat'],
  path: string,
  index: number,
) => {
  if (!path) {
    throw new FetchError("invalid root data");
  }
  switch(strategy) {
    case 'bracket':
    case 'bracket-comma':
    case 'bracket-line':
      return `${path}[]`;
    case 'index':
      return `${path}[${index}]`;
    case 'dot':
      return `${path}.${index}`;
    case 'comma':
    case 'line':
    case 'default':
    default:
      return path;
  }
}

const objectKey = (
  strategy: FormatOptions['objectFormat'],
  path: string,
  key: string,
): string => {
  if (!path) {
    throw new FetchError("invalid root data");
  }
  switch(strategy) {
    case 'dot':
      return `${path}.${key}`;
    case 'colon':
      return `${path}:${key}`;
      case 'bracket':
    default:
      return `${path}[${key}]`;
  }
}

export function merge(
  merger: Merger,
  params: any,
  options: FormatOptions | undefined,
  supportedFile?: boolean,
) {
  const opts = resolveFormatOptions(options);

  const arrayShallowMode = opts.arrayFormat === 'comma'
      || opts.arrayFormat === 'line'
      || opts.arrayFormat === 'bracket-comma'
      || opts.arrayFormat === 'bracket-line';

  let append = 0;

  const withArray = ((cb: () => void) => {
    append++;
    cb();
    append--;
  });

  const walk = (value: any, path: string, report?: (s: string) => void) => {
    const use = (name: string, value: Blob | string) => {
      if (!name) {
        throw new FetchError("invalid root data");
      } else if (append > 0) {
        merger.append(name, value);
      } else {
        merger.set(name, value);
      }
    };

    const smartUse = (name: string, value: string) => {
      if (!name) {
        throw new FetchError("invalid root data");
      } else if (report && arrayShallowMode) {
        report(value);
      } else {
        use(name, value);
      }
    };

    const shallowUse = (items: string[]) => {
      // | 'comma'         // 'foo=1,2,3'
      // | 'line'          // 'foo=1|2|3'
      // | 'bracket-comma' // 'foo[]=1,2,3'
      // | 'bracket-line'; // 'foo[]=1|2|3'
      const key = opts.arrayFormat.startsWith('bracket-') ? path + '[]' : path;
      const sep = opts.arrayFormat.endsWith("-comma") ? "," : "|";
      smartUse(key, items.map(v => encodeURIComponent(v)).join(sep));
    };

    if (value == null) {
      if (!opts.skipNull) {
        smartUse(path, opts.null);
      }
    } else if (isString(value)) {
      if (value || !opts.skipEmptyString) {
        smartUse(path, value);
      }
    } else if (isNumber(value)) {
      if (!isNaN(value)) {
        smartUse(path, String(value));
      } else if (!opts.skipNaN) {
        smartUse(path, opts.nan);
      }
    } else if (isBoolean(value)) {
      smartUse(path, opts.bool(value));
    } else if (isDate(value)) {
      smartUse(path, opts.date(value));
    } else if (report && arrayShallowMode) {
      throw new FetchError("The data is too complicated");
    } else if (value instanceof URLSearchParams) {
      for (const key of value.keys()) {
        const name = objectKey(opts.objectFormat, path, key);
        for (const str of value.getAll(key)) {
          smartUse(name, str);
        }
      }
    } else if (value instanceof Blob) {
      if (!path) {
        throw new FetchError("invalid root data");
      } else if (!supportedFile) {
        throw new FetchError("Unsupport File type");
      } else {
        use(path, value);
      }
    } else if (value instanceof FormData) {
      value.forEach((_, key) => {
        const name = objectKey(opts.objectFormat, path, key);
        value.getAll(key).forEach((val) => {
          if (!(val instanceof File)) {
            merger.append(name, val);
          } else if (!supportedFile) {
            throw new Error("Unsupported file type.");
          } else {
            merger.append(name, val);
          }
        });
      });
    } else if (isSet(value)) {
      withArray(() => {
        const items: string[] = [];
        Array.from(value).forEach((val, index) => {
          const name = arrayKey(opts.arrayFormat, path, index);
          walk(val, name, s => items.push(s));
        });
        shallowUse(items);
      })
    } else if (isMap(value)) {
      const keys = Array.from(value.keys());
      if (keys.every(k => isNumber(k) && !isNaN(k))) {
        withArray(() => {
          const items: string[] = [];
          (value as Map<number, any>).forEach((val, index) => {
            const name = arrayKey(opts.arrayFormat, path, index);
            walk(val, name, (s) => items.push(s));
          });
          shallowUse(items);
        });
      } else if (keys.every(k => isString(k))) {
        value.forEach((val, key) => {
          const name = objectKey(opts.objectFormat, path, key);
          walk(val, name);
        });
      } else {
        throw new FetchError("invalid data of a Map type");
      }
    } else if (isArray(value)) {
      withArray(() => {
        const items: string[] = [];
        value.forEach((val, index) => {
          const name = arrayKey(opts.arrayFormat, path, index);
          walk(val, name, (s) => items.push(s));
        });
        shallowUse(items);
      });
    } else if (isObject(value)) {
      for (const [key, val] of Object.entries(value)) {
        const name = objectKey(opts.objectFormat, path, key);
        walk(val, name);
      }
    } else {
      throw new FetchError("Unsupported data type.");
    }
  }

  walk(params, "");
}
