import Alert, { AlertProps } from "@mui/joy/Alert";
import Button from "@mui/joy/Button";
import IconButton from "@mui/joy/IconButton";
import Typography from "@mui/joy/Typography";
import { styled, VariantProp } from "@mui/joy/styles";
import InfoIcon from '@mui/icons-material/Info';
import WarningIcon from '@mui/icons-material/Warning';
import ReportIcon from '@mui/icons-material/Report';
import CheckCircleIcon from '@mui/icons-material/CheckCircle';
import CloseRoundedIcon from '@mui/icons-material/CloseRounded';
import {
  NotificationPayload,
  NotificationPosition,
  NotificationType,
  useNotificationContext,
  useTranslate,
  useUndoableDispatcher
} from "@rakit/core";
import {
  Toaster,
  toast,
  ExternalToast
} from 'sonner'
import {
  Dispatch,
  ReactNode,
  SetStateAction,
  useCallback,
  useEffect,
  useMemo,
  useState
} from "react";

const colorMap: Record<NotificationType, AlertProps["color"]> = {
  success: 'success',
  info: 'primary',
  warning: 'warning',
  error: 'danger',
}

const iconMap: Record<NotificationType, ReactNode> = {
  success: <CheckCircleIcon />,
  info: <InfoIcon />,
  warning: <WarningIcon />,
  error: <ReportIcon />,
}

function getColor(type: NotificationType | undefined): AlertProps["color"] {
  return type ? colorMap[type] : "neutral";
}

function getIcon(
  type: NotificationType | undefined,
  icon: ReactNode | undefined
): ReactNode | undefined {
  return icon ?? (type ? iconMap[type] : null);
}

export interface NotificationProps {
  expand?: boolean;
  multiLine?: boolean;
  position?: NotificationPosition;
  autoHideDuration?: number;
  visibleToasts?: number;
  dismissable?: boolean;
  gap?: number;
  offset?: string | number;
  dir?: 'rtl' | 'ltr' | 'auto';
  containerAriaLabel?: string;
  variant?: VariantProp;
}

type ToastPayload = NotificationPayload & { id: number | string }

function useRenderNotification(
  defaultDismissable: boolean | undefined,
  defaultMultiLine: boolean | undefined,
  variant: VariantProp,
  setToastItems: Dispatch<SetStateAction<ToastPayload[]>>
) {
  const translate = useTranslate();
  const dispatchUndo = useUndoableDispatcher();

  return useCallback((notification: NotificationPayload) => {
    const {
      description,
      descriptionArgs,
      message,
      messageArgs,
      multiLine = defaultMultiLine,
      icon,
      type,
      undoable,
      undoableId,
      dismissable = defaultDismissable,
    } = notification;

    let isDismissed = false;
    let id: string | number;

    const onDismiss = (): void => {
      if (isDismissed) {
        return;
      }
      isDismissed = true;
      setToastItems(list => {
        const index = list.findIndex(n => n.id === id);
        if (index === -1) return list;
        return list.toSpliced(index, 1);
      });
    }

    const data: ExternalToast = {
      duration: notification.autoHideDuration,
      position: notification.position,
      unstyled: true,
      onAutoClose: onDismiss,
      onDismiss: onDismiss,
    };

    const color = getColor(type);

    const undoButton = undoable ? (
      <Button
        color={color}
        size="sm"
        sx={{ mr: dismissable ? 1 : 0 }}
        variant={variant}
        onClick={() => dispatchUndo(undoableId!, true)}
      >
        <>{translate('ra.action.undo')}</>
      </Button>
    ) : null;

    const dismissButton = dismissable ? (
      <IconButton
        variant={variant}
        color={color}
        size="sm"
        onClick={() => toast.dismiss(id)}
      >
        <CloseRoundedIcon />
      </IconButton>
    ) : null;

    id = toast((
      <Alert
        variant={variant}
        color={color}
        startDecorator={getIcon(type, icon)}
        endDecorator={
          undoable && dismissable
            ? (
              <>
                {undoButton}
                {dismissButton}
              </>
            )
            : undoable
              ? undoButton
              : dismissable
                ? dismissButton
                : null
        }
        role="toast"
      >
        <div>
          {typeof message === "string" ? (
            <AutoMultiLine multiLine={multiLine}>
              {translate(message, messageArgs)}
            </AutoMultiLine>
          ) : (
            <div>{message}</div>
          )}
          {description != null ? (
            <Typography level="body-sm" color={color}>
              {typeof description === "string" ? (
                <AutoMultiLine multiLine={multiLine}>
                  {translate(description, descriptionArgs)}
                </AutoMultiLine>
              ) : description}
            </Typography>
          ) : null}
        </div>
      </Alert >
    ), data);

    setToastItems(list => [...list, {
      id,
      ...notification,
    }]);
  }, [
    defaultDismissable,
    defaultMultiLine,
    setToastItems,
    variant,
  ]);
}

/**
 * Provides a way to show a notification.
 * @see useNotify
 *
 * @example <caption>Basic usage</caption>
 * <Notification />
 *
 * @param props The component props
 * @param {string} props.type The notification type. Defaults to 'info'.
 * @param {number} props.autoHideDuration Duration in milliseconds to wait until hiding a given notification. Defaults to 4000.
 * @param {boolean} props.multiLine Set it to `true` if the notification message should be shown in more than one line.
 */

export function Notification(props: NotificationProps) {
  const {
    dismissable: defaultDismissable,
    expand = false,
    multiLine: defaultMultiLine,
    position = "bottom-center",
    visibleToasts = 3,
    variant = "soft",
  } = props;

  const { notifications, takeNotification } = useNotificationContext();
  const [toastItems, setToastItems] = useState<ToastPayload[]>([]);

  const renderNotification = useRenderNotification(
    defaultDismissable,
    defaultMultiLine,
    variant,
    setToastItems,
  );

  const hasUndoable = useMemo(
    () => toastItems.some(n => n.undoable),
    [toastItems],
  );

  useEffect(() => {
    while (notifications.length && toastItems.length < visibleToasts) {
      const notification = takeNotification();
      if (notification) {
        renderNotification(notification);
      }
    }

    const beforeunload = (e: BeforeUnloadEvent) => {
      e.preventDefault();
      const confirmationMessage = '';
      e.returnValue = confirmationMessage;
      return confirmationMessage;
    };

    if (hasUndoable) {
      window.addEventListener('beforeunload', beforeunload);

      return () => {
        window.removeEventListener('beforeunload', beforeunload);
      };
    }
  }, [
    notifications,
    toastItems,
    visibleToasts,
    hasUndoable,
    takeNotification,
    renderNotification,
  ]);

  return (
    <Toaster
      expand={expand}
      position={position}
      duration={props.autoHideDuration}
      gap={props.gap}
      offset={props.offset}
      dir={props.dir}
      visibleToasts={visibleToasts}
      containerAriaLabel={props.containerAriaLabel}
      pauseWhenPageIsHidden={hasUndoable}
    />
  )
}

const AutoMultiLine = styled('div', {
  name: "AutoMultiLine",
  overridesResolver: (props, styles) => styles.root,
  shouldForwardProp: prop => prop !== 'multiLine',
})<{ multiLine?: boolean }>(({ multiLine }) => multiLine
  ? { whiteSpace: 'pre-wrap' }
  : {}
);
