import { ComponentType, createElement } from "react";
import { useAuthenticated } from "./useAuthenticated";
import { usePermissions } from "./usePermissions";

export interface WithPermissionsProps {
  authParams?: object;
  component: ComponentType<any>;
  location?: Location;
  staticContext?: object;
  [key: string]: any;
}

/**
 * After checking that the user is authenticated,
 * retrieves the user's permissions for a specific context.
 *
 * Useful for Route components ; used internally by Resource.
 * Use it to decorate your custom page components to require
 * a custom role. It will pass the permissions as a prop to your
 * component.
 *
 * You can set additional `authParams` at will if your authProvider
 * requires it.
 *
 * @example
 *     import { Admin, CustomRoutes, WithPermissions } from 'react-admin';
 *
 *     const Foo = ({ permissions }) => (
 *         {permissions === 'admin' ? <p>Sensitive data</p> : null}
 *         <p>Not sensitive data</p>
 *     );
 *
 *     const customRoutes = [
 *         <Route path="/foo" element={
 *             <WithPermissions
 *                  authParams={{ foo: 'bar' }}
 *                  component={({ permissions, ...props }) => <Foo permissions={permissions} {...props} />}
 *              />
 *         } />
 *     ];
 *     const App = () => (
 *         <Admin>
 *             <CustomRoutes>{customRoutes}</CustomRoutes>
 *         </Admin>
 *     );
 */
export function WithPermissions(props: WithPermissionsProps) {
  const {
    authParams,
    component,
    staticContext: _,
    ...rest
  } = props;

  useAuthenticated(authParams);

  const { permissions } = usePermissions(authParams);

  // render even though the usePermissions() call isn't finished (optimistic rendering)
  return createElement(component, {
    permissions,
    ...rest
  });
}
