import { UseQueryOptions } from '@tanstack/react-query';
import { useAuthState } from "./useAuthState";

export type UseAuthenticatedOptions<ParamsType> =
  & Omit<UseQueryOptions<boolean, any>, 'queryKey' | 'queryFn'>
  & { params?: ParamsType };

/**
 * Restrict access to authenticated users.
 * Redirect anonymous users to the login page.
 *
 * Use it in your custom page components to require
 * authentication.
 *
 * You can set additional `authParams` at will if your authProvider
 * requires it.
 *
 * @example
 *     import { Admin, CustomRoutes, useAuthenticated } from 'react-admin';
 *     const FooPage = () => {
 *         useAuthenticated();
 *         return <Foo />;
 *     }
 *     const customRoutes = [
 *         <Route path="/foo" element={<FooPage />} />
 *     ];
 *     const App = () => (
 *         <Admin>
 *             <CustomRoutes>{customRoutes}</CustomRoutes>
 *         </Admin>
 *     );
 */
export function useAuthenticated<TParams = any, TError = unknown>({
  params,
  ...options
}: UseAuthenticatedOptions<TParams> = {}) {
  useAuthState<TError>(params ?? {}, true, options);
}
