import { useCallback } from "react";
import { To } from "react-router-dom";
import { useNotify } from "../notification";
import { useBasename } from "../core";
import { getErrorMessage, removeDoubleSlashes } from "../util";
import { defaultAuthParams, useAuthProvider } from "./useAuthProvider";
import { useLogout } from "./useLogout";

/**
 * Check if the current user is authenticated by calling authProvider.checkAuth().
 * Logs the user out on failure.
 *
 * @param {Object} params The parameters to pass to the authProvider
 * @param {boolean} logoutOnFailure Whether the user should be logged out if the authProvider fails to authenticate them. True by default.
 * @param {string} redirectTo The login form url. Defaults to '/login'
 *
 * @return {Promise} Resolved to the authProvider response if the user passes the check, or rejected with an error otherwise
 */
export type CheckAuth = (
  params?: any,
  logoutOnFailure?: boolean,
  redirectTo?: string
) => Promise<any>;

const checkAuthWithoutAuthProvider = () => Promise.resolve();

/**
 * Get a callback for calling the authProvider.checkAuth() method.
 * In case of rejection, redirects to the login page, displays a notification,
 * and throws an error.
 *
 * This is a low level hook. See those more specialized hooks
 * for common authentication tasks, based on useCheckAuth.
 *
 * @see useAuthenticated
 * @see useAuthState
 *
 * @returns {Function} checkAuth callback
 *
 * @example
 *
 * import { useCheckAuth } from 'react-admin';
 *
 * const MyProtectedPage = () => {
 *     const checkAuth = useCheckAuth();
 *     useEffect(() => {
 *         checkAuth().catch(() => {});
 *     }, []);
 *     return <p>Private content: EZAEZEZAET</p>
 * } // tip: use useAuthenticated() hook instead
 *
 * const MyPage = () => {
 *     const checkAuth = useCheckAuth();
 *     const [authenticated, setAuthenticated] = useState(true); // optimistic auth
 *     useEffect(() => {
 *         checkAuth({}, false)
 *              .then(() => setAuthenticated(true))
 *              .catch(() => setAuthenticated(false));
 *     }, []);
 *     return authenticated ? <Bar /> : <BarNotAuthenticated />;
 * } // tip: use useAuthState() hook instead
 */
export function useCheckAuth(): CheckAuth {
  const authProvider = useAuthProvider();
  const notify = useNotify();
  const logout = useLogout();
  const basename = useBasename();
  const loginUrl = basename
    ? removeDoubleSlashes(`${basename}/${defaultAuthParams.loginUrl}`)
    : defaultAuthParams.loginUrl;

  const checkAuth = useCallback(
    (
      params: any = {},
      logoutOnFailure = true,
      redirectTo?: To
    ) => authProvider
        ? authProvider.checkAuth(params).catch(error => {
          if (logoutOnFailure) {
            logout(
              {},
              error && error.redirectTo != null
                ? error.redirectTo
                : (redirectTo ?? loginUrl)
            );
            const shouldSkipNotify = error && error.message === false;
            !shouldSkipNotify &&
              notify(
                getErrorMessage(
                  error,
                  'ra.auth.auth_check_error'
                ),
                { type: 'error' }
              );
          }
          throw error;
        })
        : checkAuthWithoutAuthProvider(),
    [authProvider, logout, notify, loginUrl]
  );

  return checkAuth;
}
