import {
    useQuery,
    UseQueryOptions,
    QueryObserverResult,
} from '@tanstack/react-query';
import { useEffect, useMemo } from "react";
import { UserIdentity } from "./types";
import { useAuthProvider } from "./useAuthProvider";
import useInvariant from '@rakit/use-invariant';

export interface UseGetIdentityOptions<ErrorType extends Error = Error>
  extends Omit<UseQueryOptions<UserIdentity, ErrorType>, 'queryKey' | 'queryFn'> {
  onSuccess?: (data: UserIdentity) => void;
  onError?: (err: Error) => void;
  onSettled?: (data?: UserIdentity, error?: Error | null) => void;
}

export type UseGetIdentityResult<ErrorType = Error> = QueryObserverResult<
  UserIdentity,
  ErrorType
> & {
  identity: UserIdentity | undefined;
};

const defaultIdentity: UserIdentity = { id: '' };

/**
 * Return the current user identity by calling authProvider.getIdentity() on mount
 *
 * The return value updates according to the call state:
 *
 * - mount: { isPending: true }
 * - success: { identity, refetch: () => {}, isPending: false }
 * - error: { error: Error, isPending: false }
 *
 * The implementation is left to the authProvider.
 *
 * @returns The current user identity. Destructure as { isPending, identity, error, refetch }.
 *
 * @example
 * import { useGetIdentity, useGetOne } from 'react-admin';
 *
 * const PostDetail = ({ id }) => {
 *     const { data: post, isPending: postLoading } = useGetOne('posts', { id });
 *     const { identity, isPending: identityLoading } = useGetIdentity();
 *     if (postLoading || identityLoading) return <>Loading...</>;
 *     if (!post.lockedBy || post.lockedBy === identity.id) {
 *         // post isn't locked, or is locked by me
 *         return <PostEdit post={post} />
 *     } else {
 *         // post is locked by someone else and cannot be edited
 *         return <PostShow post={post} />
 *     }
 * }
 */
export function useGetIdentity<ErrorType extends Error = Error>(
  options: UseGetIdentityOptions<ErrorType> = {}
): UseGetIdentityResult<ErrorType> {
  const authProvider = useAuthProvider();
  const { onSuccess, onError, onSettled, ...queryOptions } = options;

  const result = useQuery({
    queryKey: ['auth', 'getIdentity'],
    queryFn: async ({ signal }) => {
      if (
        authProvider &&
        typeof authProvider.getIdentity === 'function'
      ) {
        return authProvider.getIdentity({ signal });
      } else {
        return defaultIdentity;
      }
    },
    ...queryOptions,
  });

  const onSuccessEvent = useInvariant(onSuccess ?? noop);
  const onErrorEvent = useInvariant(onError ?? noop);
  const onSettledEvent = useInvariant(onSettled ?? noop);

  useEffect(() => {
    if (result.data === undefined || result.isFetching) return;
    onSuccessEvent(result.data);
  }, [onSuccessEvent, result.data, result.isFetching]);

  useEffect(() => {
    if (result.error == null || result.isFetching) return;
    onErrorEvent(result.error);
  }, [onErrorEvent, result.error, result.isFetching]);

  useEffect(() => {
    if (result.status === 'pending' || result.isFetching) return;
    onSettledEvent(result.data, result.error);
  }, [
    onSettledEvent,
    result.data,
    result.error,
    result.status,
    result.isFetching,
  ]);

  return useMemo(
    () => ({
      ...result,
      identity: result.data,
    }),
    [result],
  );
}
