import { QueryClient, QueryClientProvider } from "@tanstack/react-query";
import { ReactNode, useMemo } from "react";
import { AuthContext, AuthProvider } from "../auth";
import { DataProvider, DataProviderContext, defaultDataProvider } from "../data";
import { I18nContextProvider, I18nProvider } from "../i18n";
import { NotificationContextProvider } from "../notification";
import { defaultStore, Store, StoreContextProvider } from "../store";
import { AdminRouter } from "./AdminRouter";

export interface CoreAdminContextProps {
  /**
   * The authentication provider for security and permissions
   *
   * @see https://marmelab.com/react-admin/Authentication.html
   * @example
   * import authProvider from './authProvider';
   *
   * const App = () => (
   *     <Admin authProvider={authProvider}>
   *         ...
   *     </Admin>
   * );
   */
  authProvider?: AuthProvider;

  /**
   * The base path for all URLs generated by react-admin.
   *
   * @see https://marmelab.com/react-admin/Admin.html#using-react-admin-in-a-sub-path
   * @example
   * import { Admin } from 'react-admin';
   * import { BrowserRouter } from 'react-router-dom';
   * import { dataProvider } from './dataProvider';
   *
   * const App = () => (
   *     <BrowserRouter>
   *         <Admin basename="/admin" dataProvider={dataProvider}>
   *              ...
   *         </Admin>
   *    </BrowserRouter>
   * );
   */
  basename?: string;

  children?: ReactNode;

  /**
   * The data provider used to communicate with the API
   *
   * @see https://marmelab.com/react-admin/DataProviders.html
   * @example
   * import { Admin } from 'react-admin';
   * import simpleRestProvider from 'ra-data-simple-rest';
   * const dataProvider = simpleRestProvider('http://path.to.my.api/');
   *
   * const App = () => (
   *     <Admin dataProvider={dataProvider}>
   *         ...
   *     </Admin>
   * );
   */
  dataProvider?: DataProvider;

  /**
   * The internationalization provider for translations
   *
   * @see https://marmelab.com/react-admin/Translation.html
   * @example
   * // in src/i18nProvider.js
   * import polyglotI18nProvider from 'ra-i18n-polyglot';
   * import fr from 'ra-language-french';
   *
   * export const i18nProvider = polyglotI18nProvider(() => fr, 'fr');
   *
   * // in src/App.js
   * import { Admin } from 'react-admin';
   * import { dataProvider } from './dataProvider';
   * import { i18nProvider } from './i18nProvider';
   *
   * const App = () => (
   *     <Admin dataProvider={dataProvider} i18nProvider={i18nProvider}>
   *         ...
   *     </Admin>
   * );
   */
  i18nProvider?: I18nProvider;

  /**
   * The react-query client
   *
   * @see https://marmelab.com/react-admin/Admin.html#queryclient
   * @example
   * import { Admin } from 'react-admin';
   * import { QueryClient } from '@tanstack/react-query';
   *
   * const queryClient = new QueryClient({
   *     defaultOptions: {
   *         queries: {
   *             retry: false,
   *             structuralSharing: false,
   *         },
   *         mutations: {
   *             retryDelay: 10000,
   *         },
   *     },
   * });
   *
   * const App = () => (
   *     <Admin queryClient={queryClient} dataProvider={...}>
   *         ...
   *     </Admin>
   * );
   */
  queryClient?: QueryClient;

  /**
   * The adapter for storing user preferences
   *
   * @see https://marmelab.com/react-admin/Admin.html#store
   * @example
   * import { Admin, memoryStore } from 'react-admin';
   *
   * const App = () => (
   *     <Admin dataProvider={dataProvider} store={memoryStore()}>
   *         ...
   *     </Admin>
   * );
   */
  store?: Store;
}

export function CoreAdminContext(props: CoreAdminContextProps) {
  const {
    authProvider,
    basename,
    children,
    dataProvider = defaultDataProvider,
    i18nProvider,
    queryClient,
    store = defaultStore,
  } = props;

  const finalQueryClient = useMemo(
    () => queryClient || new QueryClient(),
    [queryClient]
  );

  return (
    <AuthContext.Provider value={authProvider}>
      <DataProviderContext.Provider value={dataProvider}>
        <StoreContextProvider value={store}>
          <QueryClientProvider client={finalQueryClient}>
            <I18nContextProvider value={i18nProvider}>
              <NotificationContextProvider>
                <AdminRouter basename={basename}>
                  {children}
                </AdminRouter>
              </NotificationContextProvider>
            </I18nContextProvider>
          </QueryClientProvider>
        </StoreContextProvider>
      </DataProviderContext.Provider>
    </AuthContext.Provider>
  );
}
