import {
  ComponentType,
  useEffect,
  useState
} from 'react';
import {
  Navigate,
  Route,
  Routes,
  To
} from 'react-router-dom';
import {
  AccessControlProvider,
  LogoutOnMount,
  useCheckAuth
} from '../auth';
import { useScrollToTop } from '../scrollPosition';
import { DefaultLayout } from './DefaultLayout';
import { HasDashboardContextProvider } from './HasDashboardContextProvider';
import {
  AdminChildren,
  CoreLayoutProps
} from './types';
import { useConfigureAdminRouterFromChildren } from './useConfigureAdminRouterFromChildren';

export interface CoreAdminRoutesProps {
  /**
   * A catch-all react component to display when the URL does not match any
   *
   * @see https://marmelab.com/react-admin/Admin.html#catchall
   * @example
   * // in src/NotFound.js
   * import Card from '@mui/material/Card';
   * import CardContent from '@mui/material/CardContent';
   * import { Title } from 'react-admin';
   *
   * export const NotFound = () => (
   *     <Card>
   *         <Title title="Not Found" />
   *         <CardContent>
   *             <h1>404: Page not found</h1>
   *         </CardContent>
   *     </Card>
   * );
   *
   * // in src/App.js
   * import { Admin } from 'react-admin';
   * import { dataProvider } from './dataProvider';
   * import { NotFound } from './NotFound';
   *
   * const App = () => (
   *     <Admin catchAll={NotFound} dataProvider={dataProvider}>
   *         ...
   *     </Admin>
   * );
   */
  catchAll?: ComponentType<any> | null;

  children?: AdminChildren;

  /**
   * @example
   * import { WithPermissions } from "rwas";
   * import HomepageView from "./views/Homepage";
   *
   * const authParams = {
   *     params: { route: 'homepage' },
   * };
   *
   * const firstpage = (
   *     <WithPermissions
   *         authParams={authParams}
   *         component={HomepageView}
   *     />
   * )
   */
  dashboard?: ComponentType<any> | null;

  initialLocation?: To;

  /**
   * The main app layout component
   *
   * @see https://marmelab.com/react-admin/Admin.html#layout
   * @example
   * import { Admin, Layout } from 'react-admin';
   *
   * const MyLayout = ({ children }) => (
   *     <Layout appBarAlwaysOn>
   *        {children}
   *     </Layout>
   * );
   *
   * export const App = () => (
   *     <Admin dataProvider={dataProvider} layout={MyLayout}>
   *         ...
   *     </Admin>
   * );
   */
  layout?: ComponentType<CoreLayoutProps>;

  /**
   * The component displayed while fetching the auth provider if the admin child is an async function
   */
  loading?: ComponentType<any> | null;

  /**
   * The page to display when the admin has no Resource children
   *
   * @see https://marmelab.com/react-admin/Admin.html#ready
   * @example
   * import { Admin } from 'react-admin';
   *
   * const Ready = () => (
   *     <div>
   *         <h1>Admin ready</h1>
   *         <p>You can now add resources</p>
   *     </div>
   * )
   *
   * const App = () => (
   *     <Admin ready={Ready}>
   *         ...
   *     </Admin>
   * );
   */
  ready?: ComponentType<any> | null;

  /**
   * Flag to require authentication for all routes. Defaults to false.
   *
   * @see https://marmelab.com/react-admin/Admin.html#requireauth
   * @example
   * import { Admin } from 'react-admin';
   * import { dataProvider } from './dataProvider';
   * import { authProvider } from './authProvider';
   *
   * const App = () => (
   *     <Admin
   *         requireAuth
   *         authProvider={authProvider}
   *         dataProvider={dataProvider}
   *     >
   *         ...
   *     </Admin>
   * );
   */
  requireAuth?: boolean;
}

export function CoreAdminRoutes(props: CoreAdminRoutesProps) {
  useScrollToTop();

  const {
    routesWithLayout,
    routesWithoutLayout,
    status,
  } = useConfigureAdminRouterFromChildren(props.children);

  const {
    catchAll: CatchAll,
    dashboard: Dashboard,
    initialLocation,
    layout: Layout = DefaultLayout,
    loading: Loading,
    requireAuth,
    ready: Ready,
  } = props;

  const [onlyAnonymousRoutes, setOnlyAnonymousRoutes] = useState(requireAuth);
  const [checkAuthLoading, setCheckAuthLoading] = useState(requireAuth);
  const checkAuth = useCheckAuth();

  useEffect(() => {
    if (requireAuth) {
      // do not log the user out on failure to allow access to custom routes with no layout
      // for other routes, the LogoutOnMount component will log the user out
      checkAuth(undefined, false)
        .then(() => {
          setOnlyAnonymousRoutes(false);
        })
        .catch(() => { })
        .finally(() => {
          setCheckAuthLoading(false);
        });
    }
  }, [checkAuth, requireAuth]);

  if (status === 'empty') {
    if (!Ready) {
      throw new Error(
        'The admin is empty. Please provide an empty component, ' +
        'or pass Route or CustomRoutes as children.'
      );
    }
    return <Ready />;
  }

  if (status === 'loading' || checkAuthLoading) {
    return (
      <Routes>
        {/* Render the routes that were outside the child function. */}
        {routesWithoutLayout}
        <Route
          path="*"
          element={
            <div style={{ height: '100vh' }}>
              {Loading ? <Loading /> : null}
            </div>
          }
        />
      </Routes>
    );
  }

  if (onlyAnonymousRoutes) {
    return (
      <Routes>
        {routesWithoutLayout}
        <Route
          path="*"
          element={<LogoutOnMount />}
        />
      </Routes>
    );
  }

  return (
    <Routes>
      {routesWithoutLayout}
      <Route
        path="/*"
        element={
          <HasDashboardContextProvider value={!!Dashboard}>
            <AccessControlProvider>
              <Layout>
                <Routes>
                  {routesWithLayout}

                  {Dashboard
                    ? (<Route path="/" element={<Dashboard />} />)
                    : (initialLocation && initialLocation !== "/")
                      ? (<Route path="/" element={<Navigate to={initialLocation} />} />)
                      : null}

                  {CatchAll ? (
                    <Route
                      path="*"
                      element={<CatchAll />}
                    />
                  ) : null}
                </Routes>
              </Layout>
            </AccessControlProvider>
          </HasDashboardContextProvider>
        }
      />
    </Routes>
  );
}
