import { ComponentType, ReactElement } from "react";
import { Route, Routes, To } from "react-router-dom";
import { CoreAdminRoutes } from "./CoreAdminRoutes";
import { DefaultError } from "./DefaultError";
import { DefaultLayout } from "./DefaultLayout";
import { DefaultTitleContextProvider } from "./DefaultTitleContextProvider";
import { ErrorBoundary } from "./ErrorBoundary";
import { AdminChildren, CoreLayoutProps } from "./types";
import { DefaultReady } from "./DefaultReady";

export interface CoreAdminUIProps {
  /**
   * The content displayed when the user visits the /auth-callback page, used for redirection by third-party authentication providers
   *
   * @see https://marmelab.com/react-admin/Admin.html#authcallbackpage
   * @example
   * import { Admin } from 'react-admin';
   * import { dataProvider } from './dataProvider';
   * import { authProvider } from './authProvider';
   * import MyAuthCallbackPage from './MyAuthCallbackPage';
   *
   * const App = () => (
   *     <Admin
   *         authCallbackPage={MyAuthCallbackPage}
   *         authProvider={authProvider}
   *         dataProvider={dataProvider}
   *     >
   *         ...
   *     </Admin>
   * );
   */
  authCallbackPage?: ComponentType<any> | null;

  /**
   * A catch-all react component to display when the URL does not match any
   *
   * @see https://marmelab.com/react-admin/Admin.html#catchall
   * @example
   * // in src/NotFound.js
   * import Card from '@mui/material/Card';
   * import CardContent from '@mui/material/CardContent';
   * import { Title } from 'react-admin';
   *
   * export const NotFound = () => (
   *     <Card>
   *         <Title title="Not Found" />
   *         <CardContent>
   *             <h1>404: Page not found</h1>
   *         </CardContent>
   *     </Card>
   * );
   *
   * // in src/App.js
   * import { Admin } from 'react-admin';
   * import { dataProvider } from './dataProvider';
   * import { NotFound } from './NotFound';
   *
   * const App = () => (
   *     <Admin catchAll={NotFound} dataProvider={dataProvider}>
   *         ...
   *     </Admin>
   * );
   */
  catchAll?: ComponentType<any> | null;

  children?: AdminChildren;

  /**
   * The component to use for the dashboard page (displayed on the `/` route).
   *
   * @see https://marmelab.com/react-admin/Admin.html#dashboard
   * @example
   * import { Admin } from 'react-admin';
   * import Dashboard from './Dashboard';
   * import { dataProvider } from './dataProvider';
   *
   * const App = () => (
   *     <Admin dashboard={Dashboard} dataProvider={dataProvider}>
   *         ...
   *     </Admin>
   * );
   */
  dashboard?: ComponentType | null;

  /**
   * Set to true to disable anonymous telemetry collection
   *
   * @see https://marmelab.com/react-admin/Admin.html#disabletelemetry
   * @example
   * import { Admin } from 'react-admin';
   * import { dataProvider } from './dataProvider';
   *
   * const App = () => (
   *     <Admin disableTelemetry dataProvider={dataProvider}>
   *         ...
   *     </Admin>
   * );
   */
  disableTelemetry?: boolean;

  /**
   * The component displayed when an error is caught in a child component
   * @see https://marmelab.com/react-admin/Admin.html#error
   * @example
   * import { Admin } from 'react-admin';
   * import { MyError } from './error';
   *
   * const App = () => (
   *     <Admin error={MyError}>
   *         ...
   *     </Admin>
   * );
   */
  error?: ComponentType<any> | null;

  initialLocation?: To;

  /**
   * The main app layout component
   *
   * @see https://marmelab.com/react-admin/Admin.html#layout
   * @example
   * import { Admin, Layout } from 'react-admin';
   *
   * const MyLayout = ({ children }) => (
   *     <Layout appBarAlwaysOn>
   *        {children}
   *     </Layout>
   * );
   *
   * export const App = () => (
   *     <Admin dataProvider={dataProvider} layout={MyLayout}>
   *         ...
   *     </Admin>
   * );
   */
  layout?: ComponentType<CoreLayoutProps>;

  /**
   * The component displayed while fetching the auth provider if the admin child is an async function
   */
  loading?: ComponentType<any> | null;

  /**
   * The component displayed when the user visits the /login page
   * @see https://marmelab.com/react-admin/Admin.html#loginpage
   * @example
   * import { Admin } from 'react-admin';
   * import { dataProvider } from './dataProvider';
   * import { authProvider } from './authProvider';
   * import MyLoginPage from './MyLoginPage';
   *
   * const App = () => (
   *     <Admin
   *         loginPage={MyLoginPage}
   *         authProvider={authProvider}
   *         dataProvider={dataProvider}
   *     >
   *         ...
   *     </Admin>
   * );
   */
  loginPage?: ComponentType<any> | null;

  /**
   * The page to display when the admin has no Resource children
   *
   * @see https://marmelab.com/react-admin/Admin.html#ready
   * @example
   * import { Admin } from 'react-admin';
   *
   * const Ready = () => (
   *     <div>
   *         <h1>Admin ready</h1>
   *         <p>You can now add resources</p>
   *     </div>
   * )
   *
   * const App = () => (
   *     <Admin ready={Ready}>
   *         ...
   *     </Admin>
   * );
   */
  ready?: ComponentType<any> | null;

  /**
   * Flag to require authentication for all routes. Defaults to false.
   *
   * @see https://marmelab.com/react-admin/Admin.html#requireauth
   * @example
   * import { Admin } from 'react-admin';
   * import { dataProvider } from './dataProvider';
   * import { authProvider } from './authProvider';
   *
   * const App = () => (
   *     <Admin
   *         requireAuth
   *         authProvider={authProvider}
   *         dataProvider={dataProvider}
   *     >
   *         ...
   *     </Admin>
   * );
   */
  requireAuth?: boolean;

  /**
   * The title of the error page
   * @see https://marmelab.com/react-admin/Admin.html#title
   * @example
   * import { Admin } from 'react-admin';
   * import { dataProvider } from './dataProvider';
   *
   * const App = () => (
   *    <Admin title="My Admin" dataProvider={dataProvider}>
   *       ...
   *   </Admin>
   * );
   */
  title?: string | ReactElement | null;
}

export const CoreAdminUI = (props: CoreAdminUIProps) => {
  const {
    authCallbackPage: AuthCallbackPage,
    catchAll,
    children,
    dashboard,
    // disableTelemetry = false,
    error = DefaultError,
    initialLocation,
    layout = DefaultLayout,
    loading,
    loginPage: LoginPage,
    ready = DefaultReady,
    requireAuth = false,
    title = 'React Admin',
  } = props;

  // useEffect(() => {
  //   if (
  //     disableTelemetry ||
  //     process.env.NODE_ENV !== 'production' ||
  //     typeof window === 'undefined' ||
  //     typeof window.location === 'undefined' ||
  //     typeof Image === 'undefined'
  //   ) {
  //     return;
  //   }
  //   const img = new Image();
  //   img.src = `https://react-admin-telemetry.marmelab.com/react-admin-telemetry?domain=${window.location.hostname}`;
  // }, [disableTelemetry]);

  return (
    <DefaultTitleContextProvider value={title}>
      <ErrorBoundary error={error}>
        <Routes>
          {LoginPage != null ? (
            <Route
              path="/login"
              element={<LoginPage />}
            />
          ) : null}

          {AuthCallbackPage != null ? (
            <Route
              path="/auth-callback"
              element={<AuthCallbackPage />}
            />
          ) : null}

          <Route
            path="/*"
            element={
              <CoreAdminRoutes
                catchAll={catchAll}
                dashboard={dashboard}
                initialLocation={initialLocation}
                layout={layout}
                loading={loading}
                ready={ready}
                requireAuth={requireAuth}
              >
                {children}
              </CoreAdminRoutes>
            }
          />
        </Routes>
      </ErrorBoundary>
    </DefaultTitleContextProvider>
  );
};
