import { ComponentType, ErrorInfo, ReactElement, ReactNode, useState } from "react";
import { ErrorBoundary as ReactErrorBoundary } from "react-error-boundary";
import { getReactElement } from "../util";
import { ErrorContextProvider } from "./ErrorContextProvider";

export interface ErrorBoundaryProps {
  children?: ReactNode;

  /**
   * The component displayed when an error is caught in a child component
   * @see https://marmelab.com/react-admin/Admin.html#error
   * @example
   * import { ErrorBoundary } from 'rakit';
   * import { MyError } from './error';
   *
   * const App = () => (
   *     <ErrorBoundary error={MyError}>
   *         ...
   *     </ErrorBoundary>
   * );
   */
  error?: ComponentType<any> | ReactElement | null

  skipCaptureIfMissErrorComponent?: boolean
}

export function ErrorBoundary(props: ErrorBoundaryProps) {
  const {
    children,
    error: errorElement,
    skipCaptureIfMissErrorComponent: skipCaptureIfNoErrorComponent = false,
  } = props;

  const [errorInfo, setErrorInfo] = useState<ErrorInfo>({});

  if (errorElement == null && skipCaptureIfNoErrorComponent) {
    return null;
  }

  return (
    <ReactErrorBoundary
      onError={(_: Error, info: ErrorInfo) => setErrorInfo(info)}
      fallbackRender={({ error, resetErrorBoundary }) => (
        <ErrorContextProvider value={{ error, errorInfo, resetErrorBoundary }}>
          {errorElement != null && getReactElement(errorElement)}
        </ErrorContextProvider>
      )}
    >
      {children}
    </ReactErrorBoundary>
  );
}
