import {
  Children,
  Dispatch,
  Fragment,
  isValidElement,
  PropsWithChildren,
  ReactElement,
  ReactNode,
  SetStateAction,
  useCallback,
  useEffect,
  useState
} from "react";
import { Route } from "react-router-dom";
import { useLogout, usePermissions } from "../auth";
import { useSafeSetState } from "../util";
import { RoutesWithoutLayout } from "./RoutesWithoutLayout";
import {
  AdminChildren,
  AdminRouterStatus,
  RenderRoutesFunction
} from "./types";

function isRenderRoutesFunction(children: AdminChildren): children is RenderRoutesFunction {
  return typeof children === "function";
}

function hasIteratorProtocol(children: AdminChildren): children is Iterable<ReactNode | RenderRoutesFunction> {
  return (
    typeof Symbol !== "undefined" &&
    children != null &&
    typeof (children as any)[Symbol.iterator] === "function"
  );
}

function getRenderRoutesFunctions(children: AdminChildren): RenderRoutesFunction[] {
  if (isRenderRoutesFunction(children)) {
    return [children];
  }
  if (Array.isArray(children)) {
    return children.reduce((acc, child) => {
      if (isRenderRoutesFunction(child)) {
        acc.push(child);
      }
      return acc;
    }, []);
  }
  if (hasIteratorProtocol(children)) {
    return getRenderRoutesFunctions(Array.from(children));
  }
  return [];
}

export function useConfigureAdminRouterFromChildren(children: AdminChildren) {
  // Gather custom routes that were declared as direct children of AppRouter
  // e.g. Not returned from the child function (if any)
  // We need to know right away whether some resources were declared to correctly
  // initialize the status at the next stop
  const doLogout = useLogout();
  const { permissions, isPending } = usePermissions();
  const [routes, setRoutes, mergeRoutes] = useRoutesState(getRoutesFromNodes(children));
  const [status, setStatus] = useSafeSetState<AdminRouterStatus>(() => getStatus(children, routes));

  if (!status) {
    throw new Error('Status should be defined');
  }

  useEffect(() => {
    let funcCounts = 0;

    const onResolve = () => {
      funcCounts -= 1;
      setTimeout(() => {
        if (funcCounts <= 0) {
          setStatus(
            routes.withLayout.length > 0 || routes.withoutLayout.length > 0
              ? 'ready'
              : 'empty'
          );
        }
      })
    }

    const resolveChildFunction = async (childFunc: RenderRoutesFunction) => {
      try {
        const childrenFuncResult = childFunc(permissions);
        if ((childrenFuncResult as Promise<ReactNode>)?.then) {
          (childrenFuncResult as Promise<ReactNode>).then(
            resolvedChildren => {
              mergeRoutes(getRoutesFromNodes(resolvedChildren));
              onResolve();
            }
          );
        } else {
          mergeRoutes(getRoutesFromNodes(childrenFuncResult as ReactNode));
          onResolve();
        }
      } catch (error) {
        funcCounts -= 1;
        console.error(error);
        doLogout();
      }
    };

    const updateFromChildren = async () => {
      const functionChild = getRenderRoutesFunctions(children);
      const newRoutes = getRoutesFromNodes(children);
      setRoutes(newRoutes);
      setStatus(
        functionChild.length > 0
          ? 'loading'
          : newRoutes.withLayout.length > 0 || newRoutes.withoutLayout.length > 0
            ? 'ready'
            : 'empty'
      );

      if (functionChild.length) {
        funcCounts = functionChild.length;
        for (const fn of functionChild) {
          resolveChildFunction(fn);
        }
      } else {
        onResolve();
      }
    };

    if (!isPending) {
      updateFromChildren();
    }
  }, [
    children,
    doLogout,
    isPending,
    permissions,
    mergeRoutes,
    setRoutes,
    setStatus,
  ]);

  return {
    routesWithLayout: routes.withLayout,
    routesWithoutLayout: routes.withoutLayout,
    status,
  };
}

type Routes = {
  withLayout: ReactNode[],
  withoutLayout: ReactNode[],
}

/*
 * A hook that store the routes and resources just like setState but also provides an additional function
 * to merge new routes and resources with the existing ones.
 */
const useRoutesState = (initialState: Routes): [
  Routes,
  Dispatch<SetStateAction<Routes>>,
  (newRoutes: Routes) => void,
] => {
  const [routes, setRoutes] = useState(initialState);

  const mergeRoutes = useCallback(
    (newRouteGroups: Routes) => {
      setRoutes(previous => ({
        withLayout: previous.withLayout.concat(
          newRouteGroups.withLayout
        ),
        withoutLayout:
          previous.withoutLayout.concat(
            newRouteGroups.withoutLayout
          ),
      }));
    },
    []
  );

  return [routes, setRoutes, mergeRoutes];
};


function getStatus(children: AdminChildren, routes: Routes): AdminRouterStatus {
  return getRenderRoutesFunctions(children).length > 0
    ? 'loading'
    : routes.withLayout.length > 0 || routes.withoutLayout.length
      ? 'ready'
      : 'empty';
}

/**
 * Inspect the children and return an array of routable elements
 */
function getRoutesFromNodes(children: AdminChildren): Routes {
  const withLayout: ReactNode[] = [];
  const withoutLayout: ReactNode[] = [];

  if (isRenderRoutesFunction(children)) {
    return {
      withLayout,
      withoutLayout,
    }
  }

  if (
    !Array.isArray(children) &&
    !isValidElement(children) &&
    hasIteratorProtocol(children)
  ) {
    return getRoutesFromNodes(Array.from(children));
  }

  Children.forEach(children, node => {
    if (!isValidElement<any>(node)) {
      // Ignore non-elements. This allows people to more easily inline
      // conditionals in their route config.
      return;
    } else if (node.type === Fragment) {
      const customRoutesFromFragment = getRoutesFromNodes(node.props.children);
      withLayout.push(...customRoutesFromFragment.withLayout);
      withoutLayout.push(...customRoutesFromFragment.withoutLayout);
    } else if (node.type === Route) {
      withLayout.push(node);
    } else if (node.type === RoutesWithoutLayout) {
      const customRoutesElement = node as ReactElement<PropsWithChildren>;
      if (customRoutesElement.props.children != null) {
        withoutLayout.push(customRoutesElement.props.children);
      }
    } else if (process.env.NODE_ENV !== "production") {
      // TODO 获取 node.type 的 displayName
      const name = typeof node.type === "string"
        ? node.type
        : ((node.type as any).displayName || node.type.name);
      throw new Error(
        `[${name}]  is not a <Route> component. ` +
        `All component children of <Routes> must be a <Route> or <RoutesWithoutLayout> or <React.Fragment>`
      );
    }
  });

  return {
    withLayout,
    withoutLayout,
  };
}
