import { useCallback } from 'react';
import { To, useNavigate } from 'react-router-dom';
import { removeDoubleSlashes } from '../util';
import { useBasename } from './useBasename';

const startsWithHttpProtocol = (s: string) => /^https?:/i.test(s);

/**
 * Hook for Redirection Side Effect
 *
 * @example
 *
 * const redirect = useRedirect();
 * // redirect to list view
 * redirect('list', 'posts');
 * // redirect to edit view
 * redirect('edit', 'posts', 123);
 * // redirect to edit view with state data
 * redirect('edit', 'comments', 123, {}, { record: { post_id: record.id } });
 * // do not redirect
 * redirect(false);
 * // redirect to the result of a function
 * redirect((resource, id, data) => ...)
 */
export function useRedirect() {
  const navigate = useNavigate();
  const basename = useBasename();

  return useCallback((
    redirectTo: To | false,
    state: object = {}
  ) => {
    if (!redirectTo) {
      return;
    } else if (typeof redirectTo === 'string' &&
      startsWithHttpProtocol(redirectTo) &&
      typeof window !== 'undefined') {
      // redirection to an absolute url
      // history doesn't handle that case, so we handle it by hand
      window.location.href = redirectTo;
      return;
    } else {
      const to = basename
        ? removeDoubleSlashes(`${basename}/${redirectTo}`)
        : redirectTo;
      // redirection to an internal link
      navigate(to, {
        state:
          // We force the scrollToTop except when navigating to a list
          // where this is already done by <RestoreScrollPosition> in <Resource>
          redirectTo === 'list'
            ? state
            : { _scrollToTop: true, ...state },
      });
      return;
    }
  }, [navigate]);
}
