import { createContext } from "react";
import { AddNotificationFunction, NotificationPayload } from "./types";

export type NotificationContextValue = {
  notifications: NotificationPayload[];
  addNotification: AddNotificationFunction;
  takeNotification: () => NotificationPayload | void;
  resetNotifications: () => void;
  listenUndo: (id: string, callback: (isUndo?: boolean) => void) => void;
  dispatchUndo: (id: string, isUndo?: boolean) => void;
};

/**
 * Context for the notification state and modifiers
 *
 * @example // display notifications
 * import { useNotificationContext } from 'llads';
 *
 * const App = () => {
 *    const { notifications } = useNotificationContext();
 *    return (
 *        <ul>
 *            {notifications.map(({ message }) => (
 *                <li key={index}>{ message }</li>
 *            ))}
 *        </ul>
 *    );
 * };
 *
 * @example // reset notifications
 * import { useNotificationContext } from 'llads';
 *
 * const ResetNotificationsButton = () => {
 *    const { resetNotifications } = useNotificationContext();
 *    return (
 *        <button onClick={() => resetNotifications()}>Reset notifications</button>
 *    );
 * };
 */
export const NotificationContext = createContext<NotificationContextValue>({
  notifications: [],
  addNotification: () => { },
  takeNotification: () => { },
  resetNotifications: () => { },
  listenUndo: () => { },
  dispatchUndo: () => { },
});
