import {
  PropsWithChildren,
  useCallback,
  useMemo,
  useRef,
  useState
} from "react";
import { NotificationPayload } from "./types";
import {
  NotificationContext,
  NotificationContextValue
} from "./NotificationContext";
import { AddNotificationContext } from "./AddNotificationContext";

export const NotificationContextProvider = ({ children }: PropsWithChildren<{}>) => {
  const callbacks = useRef<Map<string, (isUndo?: boolean) => void>>();
  const [notifications, setNotifications] = useState<NotificationPayload[]>([]);

  const addNotification = useCallback((notification: NotificationPayload) => {
    setNotifications(notifications => [...notifications, notification]);
  }, []);

  const takeNotification = useCallback(() => {
    if (notifications.length === 0) return;
    const [notification, ...rest] = notifications;
    setNotifications(rest);
    return notification;
  }, [notifications]);

  const resetNotifications = useCallback(() => {
    setNotifications([]);
  }, []);

  const listenUndo = useCallback((id: string, callback: (isUndo?: boolean) => void) => {
    callbacks.current?.set(id, callback);
  }, [callbacks]);

  const dispatchUndo = useCallback((id: string, isUndo?: boolean) => {
    const callback = callbacks.current?.get(id);
    if (callback != null) {
      callbacks.current?.delete(id);
      callback(isUndo);
    }
  }, [callbacks]);

  const contextValue = useMemo<NotificationContextValue>(
    () => ({
      notifications,
      addNotification,
      takeNotification,
      resetNotifications,
      listenUndo,
      dispatchUndo,
    }),
    [notifications] // eslint-disable-line react-hooks/exhaustive-deps
  );

  // we separate the addNotification context to avoid rerendering all components
  // that depend on useNotify when a notification is dispatched
  return (
    <NotificationContext.Provider value={contextValue}>
      <AddNotificationContext.Provider value={addNotification}>
        {children}
      </AddNotificationContext.Provider>
    </NotificationContext.Provider>
  );
};
