import { ReactNode, useCallback } from 'react';
import { NotificationOptions } from './types';
import { useAddNotificationContext } from './useAddNotificationContext';
import { useUndoableId } from './useUndoableId';

/**
 * Hook for Notification Side Effect
 *
 * @example
 *
 * const notify = useNotify();
 * // simple message (info level)
 * notify('Level complete');
 * // specify level
 * notify('A problem occurred', { type: 'error' })
 * // pass arguments to the translation function
 * notify('Deleted %{count} elements', { type: 'info', messageArgs: { smart_count: 23 } })
 * // show the action as undoable in the notification
 * notify('Post renamed', { type: 'info', undoable: true })
 */
export function useNotify() {
  const addNotification = useAddNotificationContext();
  const undoableId = useUndoableId();

  return useCallback((
    message: string | ReactNode,
    options: NotificationOptions = {}
  ) => {
    if (options.undoable) {
      if (undoableId == null) {
        "You've tried to access undoable id outside <UndoableIdProducer />.\n" +
        "Please wrap your code with it first.\n"
      }
      if (undoableId === "") {
         throw new Error('Cannot notify an undoable notification while rendering.');
      }
    }
    addNotification({
      ...options,
      message,
      undoableId,
    });
  }, [addNotification]);
}
