import { ReactNode } from "react";
import { RaRecord } from "./types";
import { RecordContext } from "./RecordContext";

export interface RecordContextProviderProps<TRecord extends RaRecord = any> {
  children: ReactNode;
  representation?: string;
  record?: TRecord;
}

/**
 * Provider for the Record Context, to store a record.
 *
 * Use the useRecordContext() hook to read the context.
 * That's what the Edit and Show components do in react-admin.
 *
 * Many react-admin components read the RecordContext, including all Field
 * components.
 *
 * @example
 *
 * import { useGetOne, RecordContextProvider } from 'ra-core';
 *
 * const Show = ({ resource, id }) => {
 *     const { data } = useGetOne(resource, { id });
 *     return (
 *         <RecordContextProvider value={data}>
 *             ...
 *         </RecordContextProvider>
 *     );
 * };
 */
export function RecordContextProvider(props: RecordContextProviderProps) {
  const { children, record, representation } = props;

  if (!record) {
    return children;
  }

  return (
    <RecordContext.Provider value={{ record, representation }}>
      {children}
    </RecordContext.Provider>
  );
}
