import { RaRecord } from "./types";
import { useRecordContext } from "./useRecordContext";

export interface UseRecordParams<
  TRecord extends RaRecord = RaRecord,
> {
  record?: TRecord;
  [key: string]: any;
}

/**
 * Hook to read the record from a RecordContext.
 *
 * Must be used within a <RecordContext> such as provided by the <EditContextProvider>
 * (e.g. as a descendent of <Edit> or <EditBase>) or within a <ShowContextProvider>
 * (e.g. as a descendent of <Show> or <ShowBase>)
 *
 * @example // basic usage
 *
 * import { useRecordContext } from 'ra-core';
 *
 * const TitleField = () => {
 *     const record = useRecordContext();
 *     return <span>{record && record.title}</span>;
 * };
 *
 * @example // allow record override via props
 *
 * import { useRecord } from 'rwas';
 *
 * const TitleField = (props) => {
 *     const record = useRecord(props);
 *     return <span>{record && record.title}</span>;
 * };
 * render(<TextField record={record} />);
 *
 * @returns {RaRecord} A record object
 */
export function useRecord<TRecord extends RaRecord = RaRecord>(
  props?: UseRecordParams<TRecord>
): TRecord | undefined {
  const context = useRecordContext<TRecord>();

  return (props && props.record) || context?.record;
}
