import { useContext } from "react";
import { RaRecord, RecordToStringFunction } from "../types";
import { RecordContext, RecordContextValue } from "./RecordContext";

export interface UseRecordContextOptions<TRecord extends RaRecord> {
  record?: TRecord;
  representation?: string | RecordToStringFunction<TRecord>;
  [key: string]: any;
}

/**
 * Hook to read the record from a RecordContext.
 *
 * Must be used within a <RecordContext> such as provided by the <EditContextProvider>
 * (e.g. as a descendent of <Edit> or <EditBase>) or within a <ShowContextProvider>
 * (e.g. as a descendent of <Show> or <ShowBase>)
 *
 * @example // basic usage
 *
 * import { useRecordContext } from 'ra-core';
 *
 * const TitleField = () => {
 *     const record = useRecordContext()?.record;
 *     return <span>{record && record.title}</span>;
 * };
 *
 * @example // allow record override via props
 *
 * import { useRecordContext } from 'ra-core';
 *
 * const TitleField = (props) => {
 *     const record = useRecordContext(props);
 *     return <span>{record && record.title}</span>;
 * };
 * render(<TextField record={record} />);
 *
 * @returns {RaRecord} A record object
 */
export function useRecordContext<
  TRecord extends RaRecord = RaRecord
>(
  options?: UseRecordContextOptions<TRecord>
): RecordContextValue<TRecord> | undefined {
  const context = useContext(RecordContext);

  if (context != null) {
    return {
      record: options?.record ?? context.record,
      representation: options?.representation ?? context.representation,
    }
  }

  if (options?.record != null) {
    const { record, representation } = options;
    return { record, representation }
  }

  return undefined;
}
