import get from "lodash/get";
import { RaRecord, RecordToStringFunction } from "./types";
import { useRecordContext } from "./useRecordContext";

/**
 * Get default string representation of a record
 *
 * @example // No customization
 * const getRecordRepresentation = useGetRecordRepresentation('posts');
 * getRecordRepresentation({ id: 1, title: 'Hello' }); // => "#1"
 *
 * @example // With <Resource name="posts" recordRepresentation="title" />
 * const getRecordRepresentation = useGetRecordRepresentation('posts');
 * getRecordRepresentation({ id: 1, title: 'Hello' }); // => "Hello"
 */
export function useRecordRepresentation<TRecord extends RaRecord = RaRecord>(params?: {
  record?: TRecord;
  representation?: string | RecordToStringFunction<TRecord>;
}): string {
  const context = useRecordContext<TRecord>();
  const record = context?.record ?? params?.record;
  const representation = context?.representation ?? params?.representation;

  if (!record) {
    return '';
  }
  if (typeof representation === 'function') {
    return representation(record);
  }
  if (typeof representation === 'string') {
    return get(record, representation);
  }
  const data = record as Record<string, string>;
  if (data.name != null && data.name !== '') {
    return data.name;
  }
  if (data.title != null && data.title !== '') {
    return data.title;
  }
  if (data.label != null && data.label !== '') {
    return data.label;
  }
  if (data.reference != null && data.reference !== '') {
    return data.reference;
  }
  return `#${record.id}`;
}
