import { useEffect } from "react";
import { useLocation } from "react-router";
import { useTrackScrollPosition } from "./useTrackScrollPosition";

/**
 * A hook that tracks the scroll position and restores it when the component mounts.
 * @param storeKey The key under which to store the scroll position in the store
 * @param debounceMs The debounce time in milliseconds
 *
 * @example
 * import { useRestoreScrollPosition } from 'ra-core';
 *
 * const MyCustomPage = () => {
 *   useRestoreScrollPosition('my-list');
 *
 *   return (
 *     <div>
 *       <h1>My Custom Page</h1>
 *       <VeryLongContent />
 *     </div>
 *   );
 * };
 */
export function useRestoreScrollPosition(storeKey: string, debounceMs = 250): void {
  const [position, setPosition] = useTrackScrollPosition(storeKey, debounceMs);
  const location = useLocation();

  useEffect(() => {
    if (position != null && location.state?._scrollToTop !== true) {
      setPosition(undefined);
      window.scrollTo(0, position);
    }
    // We only want to run this effect on mount
    // eslint-disable-next-line react-hooks/exhaustive-deps
  }, []);
}
