import debounce from "lodash/debounce";
import { useEffect } from "react";
import { useStore } from "../store";

/**
 * A hook that tracks the scroll position and stores it.
 *
 * @param storeKey The key under which to store the scroll position in the store
 * @param debounceMs The debounce time in milliseconds
 *
 * @example
 * import { useTrackScrollPosition } from 'ra-core';
 *
 * const MyCustomPage = () => {
 *   useTrackScrollPosition('my-list');
 *
 *   return (
 *     <div>
 *       <h1>My Custom Page</h1>
 *       <VeryLongContent />
 *     </div>
 *   );
 * };
 */
export function useTrackScrollPosition(storeKey: string, debounceMs = 250) {
  const [position, setPosition] = useStore(storeKey);

  useEffect(() => {
    if (typeof window === 'undefined') {
      return;
    }

    const handleScroll = debounce(() => {
      setPosition(window.scrollY);
    }, debounceMs);

    window.addEventListener('scroll', handleScroll);

    return () => {
      window.removeEventListener('scroll', handleScroll);
    };
  }, [debounceMs, setPosition]);

  return [position, setPosition];
}
