import useInvariant from "@rakit/use-invariant";
import {
  useCallback,
  useEffect,
  useRef,
  useState,
} from "react";
import {
  AsyncExecutor,
  ErrorCallback,
  ExecuteFunction,
  ExecuteOptions,
  ExecuteResult,
  RetryOptions,
  RetryResolver,
  SettledCallback,
  SuccessCallback,
  ThrowOnError,
} from "./types";

export interface UseAsyncOptions<TData, TVars, TError> {
  executor: AsyncExecutor<TData, TVars>;
  immediate?: boolean;
  onError?: ErrorCallback<TVars, TError>;
  onSettled?: SettledCallback<TData, TVars, TError>;
  onSuccess?: SuccessCallback<TData, TVars>;
  retry?: RetryResolver<TVars, TError> | RetryOptions | boolean;
  throwOnError?: ThrowOnError<TVars, TError> | boolean;
  variables?: Partial<TVars>;
}

export interface UseAsyncResult<TData, TVars, TError> {
  abort: () => void;
  data: TData | undefined;
  error: TError | undefined;
  execute: ExecuteFunction<TData, TVars>;
  isExecuting: boolean;
}

const noop = () => {}

export function useAsync<TData, TVars, TError>(
  options: UseAsyncOptions<TData, TVars, TError>,
): UseAsyncResult<TData, TVars, TError> {
  const {
    executor,
    onSuccess = noop,
    onError = noop,
    onSettled = noop,
    retry: retryOptions = {},
    throwOnError,
    variables: defaultVariables = {},
    immediate,
  } = options;

  const [isExecuting, setExecuting] = useState(true);
  const [data, setData] = useState<TData>();
  const [error, setError] = useState<TError>();

  const onSuccessEvent = useInvariant(onSuccess);
  const onErrorEvent = useInvariant(onError);
  const onSettledEvent = useInvariant(onSettled);

  const controller = useRef<AbortController>();

  const abort = (): void => {
    const ctrl = controller.current;
    if (ctrl && !ctrl.signal.aborted) {
      ctrl.abort();
    }
  };

  const execute = useCallback((
    options: ExecuteOptions<Partial<TVars>> = {}
  ): ExecuteResult<TData> => {
    const { signal, discard, variables } = options;

    abort();
    setExecuting(true);
    setError(undefined);
    setData(undefined);

    const ctrl = new AbortController();
    const onAbort = () => ctrl.abort(signal?.reason);

    signal?.addEventListener("abort", onAbort);

    controller.current = ctrl;

    // @ts-ignore
    const opts: AsyncExecutorOptions<TVars> = {
      ...defaultVariables,
      ...variables,
      signal: signal ?? ctrl.signal,
    };

    const onFinish = async (
      cb: () => Promise<TData> | TData,
    ): Promise<TData | undefined> => {
      signal?.removeEventListener("abort", onAbort);
      if (controller.current === ctrl) {
        controller.current = undefined;
      }
      if (discard?.()) {
        return;
      }
      return cb();
    }

    const handleError = (error: TError) => {
      return onFinish(() => {
        setError(error);
        onErrorEvent(error, opts);
        onSettledEvent(data, error, opts);
        setExecuting(false);
        if (throwOnError === true) {
          throw error;
        }
        if (typeof throwOnError === "function" && throwOnError(error, opts)) {
          throw error;
        }
        return Promise.reject<TData>(error);
      })
    }

    const retry = (
      error: TError,
      failureCount: number,
      retryOptions: RetryOptions | RetryResolver<TVars, TError> | boolean,
    ): Promise<TData | undefined> => {
      if (ctrl.signal.aborted || retryOptions === false) {
        return handleError(error);
      }
      if (retryOptions === true) {
        return retry(
          error,
          failureCount,
          { count: 3, delay: 50 },
        );
      }
      if (typeof retryOptions === "function") {
        return retry(
          error,
          failureCount,
          retryOptions(failureCount, error, opts),
        );
      }
      if ((retryOptions.count ?? 3) >= failureCount) {
        return handleError(error);
      }
      if ((retryOptions.delay ?? -1) > -1) {
        return new Promise((resolve, reject) => {
          setTimeout(
            () => exec(failureCount + 1).then(resolve, reject),
            retryOptions.delay ?? 0,
          );
        });
      }
      return exec(failureCount + 1);
    }

    const exec = async (failureCount: number) => {
      return Promise.resolve()
        .then(async () => await executor(opts))
        .then((value) => onFinish(() => {
          setData(value);
          onSuccessEvent(value, opts);
          onSettledEvent(value, undefined, opts);
          setExecuting(false);
          return value;
        }))
        .catch(err => retry(
          err,
          failureCount,
          retryOptions,
        ));
    }

    return {
      signal: ctrl.signal,
      abort: ctrl.abort.bind(ctrl),
      promise: exec(0),
    };
  }, [
    defaultVariables,
    executor,
    retryOptions,
    throwOnError,
  ]);

  useEffect(() => {
    if (immediate) {
      execute();
      return abort;
    }
  }, []);

  return {
    abort,
    data,
    error,
    execute,
    isExecuting,
  };
}
