import { useEffect } from "react";
import { ExecuteFunction, ExecuteResult } from "./types";

export interface UseAutoExecuteOptions<TData, TVars> {
  disabled?: boolean;
  delay?: number;
  execute: ExecuteFunction<TData, TVars>;
  call?: (cb: VoidFunction) => VoidFunction | void;

  // 其它依赖项
  [key: string]: any;
}

export function useAutomatic<TData = unknown, TVars = any>(
  options: UseAutoExecuteOptions<TData, TVars>
) : void {
  const {
    disabled,
    delay = 0,
    execute,
    call = cb => cb(),
    ...deps
  } = options;

  useEffect(() => {
    if (disabled) {
      return;
    }

    let isEffectRunning = true;
    let ctrl: ExecuteResult<TData> | undefined;

    const cancel = call(async () => {
      if (delay > -1) {
        await new Promise(resolve => setTimeout(
          resolve,
          delay,
        ));
      }
      if (!isEffectRunning) {
        return;
      }
      ctrl = execute({
        discard: () => !isEffectRunning,
      });
    });

    return () => {
      if (ctrl && !ctrl.signal.aborted) {
        ctrl.abort();
      }
      isEffectRunning = false;
      cancel?.();
    };
  }, [
    disabled,
    delay,
    execute,
    call,
    deps,
  ]);
}
