import { useCallback, useMemo, useRef } from "react";
import {
  FetchContext,
  FetchContextValue,
  FetchPublisher,
  FetchSubscriber,
} from "./FetchContext";
import { FetchKey, HttpMethod } from "./types";
import { useFetchContext } from "./useFetchContext";

export interface FetchContextProviderProps extends Omit<FetchContextValue, 'subscribe' | 'publish'> {
  inherit?: boolean;
  children?: React.ReactNode;
}

type Subscribers = Record<FetchKey, Array<(method: HttpMethod, data?: any) => void>>

let timer: ReturnType<typeof setTimeout> | undefined;

export function FetchContextProvider(props: FetchContextProviderProps) {
  const {
    children,
    inherit,
    ...localContext
  } = props;

  const parentContext = useFetchContext();
  const subscribers = useRef<Subscribers>({});

  const unsubscribe = useCallback((
    key: FetchKey,
    handler: (method: HttpMethod, data?: any) => void,
  ) => {
    if (subscribers.current[key]?.includes(handler)) {
      const subs = { ...subscribers.current };
      const index = subs[key].indexOf(handler);
      subs[key] = subs[key].toSpliced(index, 1);
      subscribers.current = subs;
    }
  }, [subscribers.current]);

  const subscribe: FetchSubscriber = React.useCallback((
    key: FetchKey,
    handler: (method: HttpMethod, data?: any) => void,
  ) => {
    const subs = { ...subscribers.current }
    subs[key] ??= [];
    subs[key].push(handler);
    subscribers.current = subs;
    const cancel = parentContext?.subscribe(key, handler);
    return () => {
      cancel?.();
      unsubscribe(key, handler);
    };
  }, [
    parentContext?.subscribe,
    subscribers.current,
    unsubscribe,
  ]);

  const publish: FetchPublisher = useCallback((
    key: FetchKey,
    method: HttpMethod,
    data?: any,
  ): void => {
    //manual debounce
    if (timer) {
      // side effects already triggered in this tick, exit
      return;
    }
    timer = setTimeout(() => {
      timer = undefined;
      parentContext?.publish(key, method, data);
    }, 0);
    subscribers.current[key]?.slice().forEach(handler => {
      handler(method, data);
    });
  }, [
    parentContext,
    subscribers.current,
  ]);

  const value = useMemo<FetchContextValue>(() => {
    if (inherit) {
      return {
        subscribe,
        publish,
        ...parentContext,
        ...localContext,
      };
    }
    return {
      ...localContext,
      subscribe,
      publish,
    };
  }, [
    inherit,
    localContext,
    parentContext,
    subscribe,
    publish,
  ]);

  return (
    <FetchContext.Provider value={value}>
      {children}
    </FetchContext.Provider>
  )
}
