import {
    ErrorCallback,
    ExecuteFunction,
    RetryOptions,
    RetryResolver,
    SettledCallback,
    SuccessCallback,
    ThrowOnError,
    WithSignal,
    useAsync,
} from "@rakit/use-async";
import { useCallback } from "react";
import { Combination } from "./FetchContext";
import { FetchError } from "./FetchError";
import { BodyParams, FetchConfig, FetchKey, SearchParams } from "./types";
import { useCreateBody } from "./useCreateBody";
import { useCreateHeaders } from "./useCreateHeaders";
import { RequestFactory, useCreateRequest } from "./useCreateRequest";
import { useCreateUrl } from "./useCreateUrl";
import { ErrorCorrector, useErrorCorrector } from "./useErrorCorrector";
import { useFetchConfig } from "./useFetchConfig";
import { Fetcher, useFetcher } from "./useFetcher";
import { useFetchPublisher } from "./useFetchPublisher";
import { RequestIntercepter, useRequestIntercepter } from "./useRequestIntercepter";
import { ResponseTransformer, useResponseTransformer } from "./useResponseTransformer";

export type FetchVariables = {
  body?: BodyParams;
  headers?: HeadersInit;
  searchParams?: SearchParams;
  hasFiles?: boolean;
}

class InternalData<TData = any> {
  constructor(
    public readonly request: Request,
    public readonly data: TData,
  ) { }
}

export interface UseFetchOptions<TData, TError> {
  key?: FetchKey;
  url: string;
  config?: FetchConfig;
  fetcher?: Fetcher;
  baseUrl?: string;
  headers?: HeadersInit;
  headersCombination?: Combination;
  searchParams?: SearchParams;
  searchParamsCombination?: Combination;
  body?: BodyParams;
  bodyCombination?: Combination;
  hasFiles?: boolean;
  createRequest?: RequestFactory;
  intercept?: RequestIntercepter;
  transform?: ResponseTransformer<TData>;
  correct?: ErrorCorrector<TData, TError | FetchError>;
  onError?: ErrorCallback<FetchVariables, TError | FetchError>;
  onSettled?: SettledCallback<TData, FetchVariables, TError | FetchError>;
  onSuccess?: SuccessCallback<TData, FetchVariables>;
  retry?: RetryResolver<FetchVariables, TError | FetchError> | RetryOptions | boolean;
  throwOnError?: ThrowOnError<FetchVariables, TError | FetchError> | boolean;
}

export interface UseFetchResult<TData, TError> {
  abort: () => void;
  data: TData | undefined;
  error: TError | FetchError | undefined;
  execFetch: ExecuteFunction<TData, FetchVariables>;
  isFetching: boolean;
}

export function useFetch<TData = any, TError = unknown>(
  options: UseFetchOptions<TData, TError>
): UseFetchResult<TData, TError> {
  const {
    key,
    url,
    onSuccess: successCallback,
    onError,
    onSettled,
    retry,
    throwOnError,
  } = options;

  const config = useFetchConfig(options);
  const createUrl = useCreateUrl(options);
  const createHeaders = useCreateHeaders(options);
  const createBody = useCreateBody(options);
  const createRequest = useCreateRequest(options);
  const fetch = useFetcher(options);
  const intercept = useRequestIntercepter(options);
  const transform = useResponseTransformer<TData>(options);
  const correct = useErrorCorrector<TData, TError>(options);
  const publish = useFetchPublisher(key);

  const onSuccess = useCallback((
    data: TData,
    opts: WithSignal<FetchVariables>
  ): void => {
    if (data instanceof InternalData) {
      successCallback?.(data.data, opts);
      publish(data.request.method ?? 'GET', data);
    } else {
      successCallback?.(data, opts);
    }
  }, [successCallback, publish]);

  const executor = useCallback(
    async (options: WithSignal<FetchVariables>): Promise<TData> => {
      const { signal, ...variables } = options;

      let uri: URL | undefined;
      let body: BodyInit | null | undefined;
      let headers: HeadersInit | undefined;
      let request: Request | undefined;
      let response: Response | undefined;

      return Promise.resolve()
        .then(() => uri = createUrl(url, variables?.searchParams))
        .then(() => body = createBody(variables?.body, variables?.hasFiles))
        .then(() => headers = createHeaders(variables?.headers))
        .then(() => createRequest(uri!, config, headers, body, signal))
        .then(req => intercept(request = req))
        .then(req => fetch(request = req))
        .then(res => transform(request!, response = res))
        // 我们使用 InternalData 代理结果
        .then(data => new InternalData(request!, data) as TData)
        .catch(async err => await correct(err, request, response));
    },
    [
      url,
      createUrl,
      createBody,
      createHeaders,
      createRequest,
      config,
      intercept,
      transform,
      correct,
    ],
  );

  const {
    abort,
    data,
    error,
    execute: execFetch,
    isExecuting: isFetching,
  } = useAsync<TData, FetchVariables, TError | FetchError>({
    executor,
    onError,
    onSettled,
    onSuccess,
    retry,
    throwOnError,
  });

  return {
    abort,
    data,
    execFetch,
    error,
    isFetching,
  };
}
